"""
Assertions for Airtest
"""

from airtestProject.airtest.core.helper import logwrap
from airtestProject.airtest.core.cv import loop_find
from airtestProject.airtest.core.error import TargetNotFoundError
from airtestProject.airtest.core.settings import Settings as ST


@logwrap
def assert_exists(v, msg=""):
    """
    Assert target exists on device screen

    :param v: target to be checked
    :param msg: short description of assertion, it will be recorded in the report
    :raise AssertionError: if assertion fails
    :return: coordinates of the target
    :platforms: Android, Windows, iOS
    :Example:

        >>> assert_exists(Template(r"tpl1607324047907.png"), "assert exists")

    """
    try:
        pos = loop_find(v, timeout=ST.FIND_TIMEOUT, threshold=ST.THRESHOLD_STRICT or v.threshold)
        return pos
    except TargetNotFoundError:
        raise AssertionError("%s does not exist in screen, message: %s" % (v, msg))


@logwrap
def assert_not_exists(v, msg=""):
    """
    Assert target does not exist on device screen

    :param v: target to be checked
    :param msg: short description of assertion, it will be recorded in the report
    :raise AssertionError: if assertion fails
    :return: None.
    :platforms: Android, Windows, iOS
    :Example:

        >>> assert_not_exists(Template(r"tpl1607324047907.png"), "assert not exists")
    """
    try:
        pos = loop_find(v, timeout=ST.FIND_TIMEOUT_TMP)
        raise AssertionError("%s exists unexpectedly at pos: %s, message: %s" % (v, pos, msg))
    except TargetNotFoundError:
        pass


@logwrap
def assert_equal(first, second, msg="", snapshot=True):  # noqa
    """
    Assert two values are equal

    :param first: first value
    :param second: second value
    :param msg: short description of assertion, it will be recorded in the report
    :raise AssertionError: if assertion fails
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        >>> assert_equal(1, 1, msg="assert 1==1")
    """
    assert first == second, "%s and %s are not equal, message: %s" % (first, second, msg)


@logwrap
def assert_not_equal(first, second, msg="", snapshot=True):  # noqa
    """
    Assert two values are not equal

    :param first: first value
    :param second: second value
    :param msg: short description of assertion, it will be recorded in the report
    :raise AssertionError: if assertion
    :return: None
    :platforms: Android, Windows, iOS
    :Example:

        >>> assert_not_equal(1, 2, msg="assert 1!=2")
    """
    assert first != second, "%s and %s are equal, message: %s" % (first, second, msg)


@logwrap
def assert_true(expr, msg="", snapshot=True):  # noqa
    """
    Assert expression is True ( bool(expr) is True )
    Note that this is equivalent to bool(expr) is True and not to expr is True (use assertIs(expr, True) for the latter).

    :Example:

        >>> assert_true(1==1, msg="assert 1==1")
    """
    assert bool(expr), "expression is not True, message: %s" % msg


@logwrap
def assert_false(expr, msg="", snapshot=True):  # noqa
    """
    Assert expression is False ( bool(expr) is False )

    :Example:

        >>> assert_false(1==2, msg="assert 1!=2")
    """
    assert not bool(expr), "expression is not False, message: %s" % msg


@logwrap
def assert_is(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first and second are the same object.

    :Example:

        >>> assert_is(1, 1, msg="assert 1 is 1")
    """
    assert first is second, "%s and %s are not the same object, message: %s" % (first, second, msg)


@logwrap
def assert_is_not(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first and second are not the same object.

    :Example:

        >>> assert_is_not(1, 2, msg="assert 1 is not 2")
    """
    assert first is not second, "%s and %s are the same object, message: %s" % (first, second, msg)


@logwrap
def assert_is_none(expr, msg="", snapshot=True):  # noqa
    """
    Test that expr is None.

    :Example:

        >>> assert_is_none(None, msg="assert None is None")
    """
    assert expr is None, "%s is not None, message: %s" % (expr, msg)


@logwrap
def assert_is_not_none(expr, msg="", snapshot=True):  # noqa
    """
    Test that expr is not None.

    :Example:

        >>> assert_is_not_none(1, msg="assert 1 is not None")
    """
    assert expr is not None, "%s is None, message: %s" % (expr, msg)


@logwrap
def assert_in(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is in second.

    :Example:

        >>> assert_in(1, [1, 2], msg="assert 1 in [1, 2]")
    """
    assert first in second, "%s is not in %s, message: %s" % (first, second, msg)


@logwrap
def assert_not_in(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is not in second.

    :Example:

        >>> assert_not_in(3, [1, 2], msg="assert 3 not in [1, 2]")
    """
    assert first not in second, "%s is in %s, message: %s" % (first, second, msg)


@logwrap
def assert_is_instance(obj, cls, msg="", snapshot=True):  # noqa
    """
    Test that obj is an instance of cls (which can be a class or a tuple of classes, as supported by isinstance()).

    :Example:

        >>> assert_is_instance(1, int, msg="assert 1 is int")
    """
    assert isinstance(obj, cls), "%s is not an instance of %s, message: %s" % (obj, cls, msg)


@logwrap
def assert_not_is_instance(obj, cls, msg="", snapshot=True):  # noqa
    """
    Test that obj is not an instance of cls.

    :Example:

        >>> assert_not_is_instance(1, str, msg="assert 1 is not str")
    """
    assert not isinstance(obj, cls), "%s is an instance of %s, message: %s" % (obj, cls, msg)


@logwrap
def assert_greater(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is greater than second. (first > second)

    :Example:

        >>> assert_greater(2, 1, msg="assert 2 > 1")
    """
    assert first > second, "%s is not greater than %s, message: %s" % (first, second, msg)


@logwrap
def assert_greater_equal(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is greater than or equal to second. (first >= second)

    :Example:

        >>> assert_greater_equal(1, 1, msg="assert 1 >= 1")
    """
    assert first >= second, "%s is not greater than or equal to %s, message: %s" % (first, second, msg)


@logwrap
def assert_less(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is less than second. (first < second)

    :Example:

        >>> assert_less(1, 2, msg="assert 1 < 2")
    """
    assert first < second, "%s is not less than %s, message: %s" % (first, second, msg)


@logwrap
def assert_less_equal(first, second, msg="", snapshot=True):  # noqa
    """
    Test that first is less than or equal to second. (first <= second)

    :Example:

        >>> assert_less_equal(1, 1, msg="assert 1 <= 1")
    """
    assert first <= second, "%s is not less than or equal to %s, message: %s" % (first, second, msg)



# encoding=utf-8
"""
MIT License

Copyright (c) 2021 zackees

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

    refer to: https://github.com/zackees/static_ffmpeg

    Entry point for running the ffmpeg executable.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import subprocess
import os
import stat
import sys
import zipfile
from datetime import datetime

import requests  # type: ignore
from filelock import FileLock, Timeout

TIMEOUT = 10 * 60  # Wait upto 10 minutes to validate install
# otherwise break the lock and install anyway.

SELF_DIR = os.path.abspath(os.path.dirname(__file__))
LOCK_FILE = os.path.join(SELF_DIR, "lock.file")


PLATFORM_ZIP_FILES = {
    "win32": "/raw/main/v5.0/win32.zip",
    "darwin": "/raw/main/v5.0/darwin.zip",
    "linux": "/raw/main/v5.0/linux.zip",
}

BACKUP_PLATFORM_ZIP_FILES = {
    "win32": "/downloads/ffmpeg_bins/v5.0/win32.zip",
    "darwin": "/downloads/ffmpeg_bins/v5.0/darwin.zip",
    "linux": "/downloads/ffmpeg_bins/v5.0/linux.zip",
}

def check_system():
    """Friendly error if there's a problem with the system configuration."""
    if sys.platform not in PLATFORM_ZIP_FILES:
        raise OSError("Please implement static_ffmpeg for %s"%sys.platform)

def get_platform_http_zip():
    """Return the download link for the current platform"""
    check_system()
    try:
        MAIN_SIT = "https://github.com/zackees/ffmpeg_bins"
        r = requests.get(MAIN_SIT, timeout=10)
        r.raise_for_status()
        return MAIN_SIT + PLATFORM_ZIP_FILES[sys.platform]
    except:
        print("Warning, main url download fail, try backup url.")
        BACKUP_SIT = "https://airtestproject.s3.netease.com"
        return BACKUP_SIT + BACKUP_PLATFORM_ZIP_FILES[sys.platform]
    
def get_platform_dir():
    """Either get the executable or raise an error"""
    check_system()
    return os.path.join(SELF_DIR, "bin", sys.platform)


def download_file(url, local_path):
    """Downloads a file to the give path."""
    # NOTE the stream=True parameter below
    print("Downloading and save to %s" % local_path)
    with requests.get(url, stream=True) as req:
        req.raise_for_status()
        with open(local_path, "wb") as file_d:
            for chunk in req.iter_content(chunk_size=8192 * 16):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                # sys.stdout.write(".")
                file_d.write(chunk)
        print("Download completed.")
    return local_path


def get_or_fetch_platform_executables_else_raise(fix_permissions=True):
    """Either get the executable or raise an error"""
    lock = FileLock(LOCK_FILE, timeout=TIMEOUT)  # pylint: disable=E0110
    try:
        with lock.acquire():
            return _get_or_fetch_platform_executables_else_raise_no_lock(
                fix_permissions=fix_permissions
            )
    except Timeout:
        sys.stderr.write(
            "%s: Warning, could not acquire lock at %s\n"
        )%(__file__, LOCK_FILE)
        return _get_or_fetch_platform_executables_else_raise_no_lock(
            fix_permissions=fix_permissions
        )


def _get_or_fetch_platform_executables_else_raise_no_lock(fix_permissions=True):
    """Either get the executable or raise an error, internal api"""
    exe_dir = get_platform_dir()
    installed_crumb = os.path.join(exe_dir, "installed.crumb")
    if not os.path.exists(installed_crumb):
        print("\n===========ffmpeg is missing, fetching it now.=============\n")
        # All zip files store their platform executables in a folder
        # like "win32" or "darwin" or "linux" inside the executable. So root
        # the install one level up from that same directory.
        install_dir = os.path.dirname(exe_dir)
        try:
            os.makedirs(exe_dir)#, exist_ok=True)
        except:
            pass
        local_zip = exe_dir + ".zip"
        url = get_platform_http_zip()
        download_file(url, local_zip)
        print("Extracting %s -> %s"%(local_zip, install_dir))
        with zipfile.ZipFile(local_zip, mode="r") as zipf:
            zipf.extractall(install_dir)
        try:
            os.remove(local_zip)
        except OSError as err:
            print("%s: Error could not remove %s because of %s"%(__file__, local_zip, err))
        with open(installed_crumb, "wt") as filed:  # pylint: disable=W1514
            filed.write("installed from %s on %s"%(url, datetime.now().__str__()))
        print("=============================================================\n")
    ffmpeg_exe = os.path.join(exe_dir, "ffmpeg")
    ffprobe_exe = os.path.join(exe_dir, "ffprobe")
    if sys.platform == "win32":
        ffmpeg_exe = "%s.exe"%ffmpeg_exe
        ffprobe_exe = "%s.exe"%ffprobe_exe
    for exe in [ffmpeg_exe, ffprobe_exe]:
        if (
            fix_permissions
            and sys.platform != "win32"
            and (not os.access(exe, os.X_OK) or not os.access(exe, os.R_OK))
        ):
            # Set bits for execution and read for all users.
            exe_bits = stat.S_IXOTH | stat.S_IXUSR | stat.S_IXGRP
            read_bits = stat.S_IRUSR | stat.S_IRGRP | stat.S_IXGRP
            os.chmod(exe, exe_bits | read_bits)
            assert os.access(exe, os.X_OK), "Could not execute %s"%exe
            assert os.access(exe, os.R_OK), "Could not get read bits of %s"%exe
    return ffmpeg_exe, ffprobe_exe


def main_static_ffmpeg():
    """Entry point for running static_ffmpeg, which delegates to ffmpeg."""
    ffmpeg_exe, _ = get_or_fetch_platform_executables_else_raise()
    rtn = subprocess.call([ffmpeg_exe] + sys.argv[1:])
    sys.exit(rtn)


def main_static_ffprobe():
    """Entry point for running static_ffmpeg, which delegates to ffmpeg."""
    _, ffprobe = get_or_fetch_platform_executables_else_raise()
    rtn = subprocess.call([ffprobe] + sys.argv[1:])
    sys.exit(rtn)


def main_print_paths():
    """Entry point for printing ffmpeg paths"""
    ffmpeg_exe, ffprobe_exe = get_or_fetch_platform_executables_else_raise()
    print("FFMPEG=%s"%ffmpeg_exe)
    print("FFPROBE=%s"%ffprobe_exe)
    sys.exit(0)

def add_paths():
    """Add the ffmpeg executable to the path"""
    ffmpeg, _ = get_or_fetch_platform_executables_else_raise()
    os.environ["PATH"] = os.pathsep.join([str(os.path.dirname(ffmpeg)), os.environ["PATH"]])

if __name__ == "__main__":
    # get_or_fetch_platform_executables_else_raise()
    add_paths()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import numpy
import socket
import traceback
from airtestProject.airtest import aircv
from airtestProject.airtest.utils.snippet import reg_cleanup, on_method_ready, ready_method
from airtestProject.airtest.core.ios.constant import ROTATION_MODE, DEFAULT_MJPEG_PORT
from airtestProject.airtest.utils.logger import get_logger
from airtestProject.airtest.utils.safesocket import SafeSocket


LOGGING = get_logger(__name__)


class SocketBuffer(SafeSocket):
    def __init__(self, sock: socket.socket):
        super(SocketBuffer, self).__init__(sock)

    def _drain(self):
        _data = self.sock.recv(1024)
        if _data is None or _data == b"":
            raise IOError("socket closed")
        self.buf += _data
        return len(_data)

    def read_until(self, delimeter: bytes) -> bytes:
        """ return without delimeter """
        while True:
            index = self.buf.find(delimeter)
            if index != -1:
                _return = self.buf[:index]
                self.buf = self.buf[index + len(delimeter):]
                return _return
            self._drain()

    def read_bytes(self, length: int) -> bytes:
        while length > len(self.buf):
            self._drain()

        _return, self.buf = self.buf[:length], self.buf[length:]
        return _return

    def write(self, data: bytes):
        return self.sock.sendall(data)


class MJpegcap(object):

    def __init__(self, instruct_helper=None, ip='localhost', port=None, ori_function=None):
        self.instruct_helper = instruct_helper
        self.port = int(port or DEFAULT_MJPEG_PORT)
        self.ip = ip
        # 如果指定了port，说明已经将wda的9100端口映射到了新端口，无需本地重复映射
        self.port_forwarding = True if self.port == DEFAULT_MJPEG_PORT and ip in ('localhost', '127.0.0.1') else False
        self.ori_function = ori_function
        self.sock = None
        self.buf = None
        self._is_running = False

    @ready_method
    def setup_stream_server(self):
        if self.port_forwarding:
            self.port, _ = self.instruct_helper.setup_proxy(9100)
        self.init_sock()
        reg_cleanup(self.teardown_stream)

    def init_sock(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((self.ip, self.port))
            self.buf = SocketBuffer(self.sock)
            self.buf.write(b"GET / HTTP/1.0\r\nHost: localhost\r\n\r\n")
            self.buf.read_until(b'\r\n\r\n')
            self._is_running = True
            LOGGING.info("mjpegsock is ready")
        except ConnectionResetError:
            # 断开tidevice或是拔线，会导致这个异常，直接退出即可
            LOGGING.error("mjpegsock connection error")
            raise

    @on_method_ready('setup_stream_server')
    def get_frame_from_stream(self):
        if self._is_running is False:
            self.init_sock()
        try:
            while True:
                line = self.buf.read_until(b'\r\n')
                if line.startswith(b"Content-Length"):
                    length = int(line.decode('utf-8').split(": ")[1])
                    break
            while True:
                if self.buf.read_until(b'\r\n') == b'':
                    break
            imdata = self.buf.read_bytes(length)
            return imdata
        except IOError:
            # 如果暂停获取mjpegsock的数据一段时间，可能会导致它断开，这里将self.buf关闭并临时返回黑屏图像
            # 等待下一次需要获取屏幕时，再进行重连
            LOGGING.debug("mjpegsock is closed")
            self._is_running = False
            self.buf.close()
            return self.get_blank_screen()

    def get_frame(self):
        # 获得单张屏幕截图
        return self.get_frame_from_stream()

    def snapshot(self, ensure_orientation=True, *args, **kwargs):
        """
        Take a screenshot and convert it into a cv2 image object

        获取一张屏幕截图，并转化成cv2的图像对象
        !!! 注意，该方法拿到的截图可能不是队列中最新的，除非一直在消费队列中的图像，否则可能会是过往图像内容，请谨慎使用

        Args:
            ensure_orientation: True or False whether to keep the orientation same as display

        Returns: numpy.ndarray

        """
        screen = self.get_frame_from_stream()
        try:
            screen = aircv.utils.string_2_img(screen)
        except Exception:
            # may be black/locked screen or other reason, print exc for debugging
            traceback.print_exc()
            return None

        if ensure_orientation:
            if self.ori_function:
                display_info = self.ori_function()
                orientation = next(key for key, value in ROTATION_MODE.items() if value == display_info["orientation"])
                screen = aircv.rotate(screen, -orientation, clockwise=False)

        return screen

    def get_blank_screen(self):
        """
        生成一个黑屏图像，在连接失效时代替屏幕画面返回
        Returns:

        """
        if self.ori_function:
            display_info = self.ori_function()
            width, height = display_info['width'], display_info['height']
            if display_info["orientation"] in [90, 270]:
                width, height = height, width
        else:
            width, height = 1080, 1920
        img = numpy.zeros((width, height, 3)).astype('uint8')
        img_string = aircv.utils.img_2_string(img)
        return img_string

    def teardown_stream(self):
        if self.buf:
            self.buf.close()
        if self.port_forwarding:
            self.instruct_helper.remove_proxy(self.port)
        self.port = None


if __name__ == "__main__":
    import wda
    from airtest.core.ios.instruct_cmd import InstructHelper
    addr = "http://localhost:8100"
    driver = wda.Client(addr)
    info = driver.info
    instruct_helper = InstructHelper(info['uuid'])
    mjpeg_server = MJpegcap(instruct_helper)
    print(len(mjpeg_server.get_frame()))